import re


def escape(string: str) -> str:
    """Escape a string to be sent through Discord, and format it using RoyalCode.

    Warning:
        Currently escapes everything, even items in code blocks."""
    url_pattern = re.compile(r"\[url=(.*?)](.*?)\[/url]")
    url_replacement = r'\2 (\1)'

    simple_parse = string \
        .replace("*", "\\*") \
        .replace("_", "\\_") \
        .replace("`", "\\`") \
        .replace("[b]", "**") \
        .replace("[/b]", "**") \
        .replace("[i]", "_") \
        .replace("[/i]", "_") \
        .replace("[u]", "__") \
        .replace("[/u]", "__") \
        .replace("[c]", "`") \
        .replace("[/c]", "`") \
        .replace("[p]", "```") \
        .replace("[/p]", "```")

    advanced_parse = re.sub(url_pattern, url_replacement, simple_parse)

    return advanced_parse



import royalnet
import royalnet.constellation.api as rca
import royalnet.utils as ru


class ApiRoyalnetVersionStar(rca.ApiStar):
    path = "/api/royalnet/version/v1"

    tags = ["royalnet"]

    @rca.magic
    async def get(self, data: rca.ApiData) -> ru.JSON:
        """Get the current Royalnet version."""
        return {
            "semantic": royalnet.__version__
        }



import logging
import multiprocessing
import time
from typing import *

import click
import toml

import royalnet.serf as rs
import royalnet.utils as ru

try:
    import royalnet.serf.telegram as rst
except ImportError:
    rst = None

try:
    import royalnet.serf.discord as rsd
except ImportError:
    rsd = None

try:
    import royalnet.constellation as rc
except ImportError:
    rc = None

try:
    import royalnet.herald as rh
except ImportError:
    rh = None

try:
    import coloredlogs
except ImportError:
    coloredlogs = None

log = logging.getLogger(__name__)


@click.command()
@click.option("-c", "--config-file", default="./config.toml", type=click.File(encoding="utf8"),
              help="The filename of the Royalnet configuration file.")
def run(config_file: str):
    # Read the configuration file
    config: dict = toml.load(config_file)

    ru.init_logging(config["Logging"])

    if config["Sentry"] is None or not config["Sentry"]["enabled"]:
        log.info("Sentry: disabled")
    else:
        try:
            ru.init_sentry(config["Sentry"])
        except ImportError:
            log.info("Sentry: not installed")
        else:
            log.info("Sentry: enabled")

    processes: Dict[str, ru.RoyalnetProcess] = {}
    """A list of all processes that the launcher should start and monitor."""

    herald_cfg = config.get("Herald")
    if rh is None:
        log.info("Herald: Not installed")
    elif herald_cfg is None:
        log.warning("Herald: Not configured")
    elif not herald_cfg["enabled"]:
        log.info("Herald: Disabled")
    elif herald_cfg["mode"] == "local":
        log.info("Herald: Enabled (local server)")

        def herald_constructor() -> multiprocessing.Process:
            # Create a Herald server
            herald_server = rh.Server(rh.Config.from_config(name="<server>", **herald_cfg))
            # Run the Herald server on a new process
            return multiprocessing.Process(
                name="Herald.Local",
                target=herald_server.run_blocking,
                daemon=True,
                kwargs={
                    "logging_cfg": config["Logging"]
                }
            )

        processes["Herald"] = ru.RoyalnetProcess(herald_constructor, None)
    elif herald_cfg["mode"] == "remote":
        log.info("Herald: Enabled (remote server)")
    else:
        log.error(f"Invalid Herald mode: {herald_cfg['mode']}")

    # Serfs
    serfs_cfg = config.get("Serfs")
    if serfs_cfg is None:
        log.warning("__serfs__: Not configured")
    else:
        log.debug("__serfs__: Configured")

        def configure_serf(n: str, module, class_: Type[rs.Serf]):
            serf_cfg = serfs_cfg.get(n)
            if module is None:
                log.info(f"Serf.{n}: Not installed")
            elif serf_cfg is None:
                log.warning(f"Serf.{n}: Not configured")
            elif not serf_cfg["enabled"]:
                log.info(f"Serf.{n}: Disabled")
            else:
                def serf_constructor() -> multiprocessing.Process:
                    return multiprocessing.Process(
                        name=f"Serf.{n}",
                        target=class_.run_process,
                        daemon=True,
                        kwargs={
                            "alchemy_cfg": config["Alchemy"],
                            "herald_cfg": herald_cfg,
                            "packs_cfg": config["Packs"],
                            "sentry_cfg": config["Sentry"],
                            "logging_cfg": config["Logging"],
                            "serf_cfg": serf_cfg,
                        }
                    )

                processes[f"Serf.{n}"] = ru.RoyalnetProcess(serf_constructor, None)
                log.info(f"Serf.{n}: Enabled")

        if rst is not None:
            configure_serf("Telegram", rst, rst.TelegramSerf)
        if rsd is not None:
            configure_serf("Discord", rsd, rsd.DiscordSerf)

    # Constellation
    constellation_cfg = config.get("Constellation")
    if rc is None:
        log.info(f"Constellation: Not installed")
    elif constellation_cfg is None:
        log.warning(f"Constellation: Not configured")
    elif not constellation_cfg["enabled"]:
        log.info(f"Constellation: Disabled")
    else:
        def constellation_constructor() -> multiprocessing.Process:
            return multiprocessing.Process(
                name="Constellation",
                target=rc.Constellation.run_process,
                daemon=True,
                kwargs={
                    "alchemy_cfg": config["Alchemy"],
                    "herald_cfg": herald_cfg,
                    "packs_cfg": config["Packs"],
                    "sentry_cfg": config["Sentry"],
                    "logging_cfg": config["Logging"],
                    "constellation_cfg": config["Constellation"],
                }
            )

        processes["Constellation"] = ru.RoyalnetProcess(constellation_constructor, None)

        log.info("Constellation: Enabled")

    try:
        # Monitor processes
        while True:
            log.debug("Checking process status...")
            for name, process in processes.items():
                if process.current_process is None:
                    log.info(f"{name}: Starting...")
                    process.current_process = process.constructor()
                    process.current_process.start()
                elif not process.current_process.is_alive():
                    log.error(f"{name}: Process is dead, restarting...")
                    process.current_process = process.constructor()
                    process.current_process.start()
            log.debug("Done, checking again in 60 seconds.")
            time.sleep(60)
    except KeyboardInterrupt:
        log.info("Received SIGTERM, stopping everything!")
        for name, process in processes.items():
            log.info(f"{name}: Killing...")
            process.current_process.kill()
        log.info("Goodbye!")


if __name__ == "__main__":
    run()



from typing import *

from .commanddata import CommandData


class KeyboardKey:
    def __init__(self,
                 short: str,
                 text: str,
                 callback: Callable[[CommandData], Awaitable[None]]):
        self.short: str = short
        self.text: str = text
        self.callback: Callable[[CommandData], Awaitable[None]] = callback

    async def press(self, data: CommandData):
        await self.callback(data)



import pkg_resources

from royalnet.commands import *


class RoyalnetversionCommand(Command):
    name: str = "royalnetversion"

    description: str = "Display the current Royalnet version."

    @property
    def royalnet_version(self):
        return pkg_resources.get_distribution("royalnet").version

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        # noinspection PyUnreachableCode
        if __debug__:
            message = f"ℹ️ Royalnet [url=https://github.com/Steffo99/royalnet/]Unreleased[/url]\n"
        else:
            message = f"ℹ️ Royalnet [url=https://github.com/Steffo99/royalnet/releases/tag/{self.royalnet_version}]" \
                      f"{self.royalnet_version}[/url]\n"
        if "69" in self.royalnet_version:
            message += "(Nice.)"
        await data.reply(message)



import dataclasses
import multiprocessing
from typing import *


@dataclasses.dataclass()
class RoyalnetProcess:
    constructor: Callable[[], multiprocessing.Process]
    current_process: Optional[multiprocessing.Process]



"""The subpackage providing all classes related to Royalnet commands."""

from .command import Command
from .commandargs import CommandArgs
from .commanddata import CommandData
from .configdict import ConfigDict
from .errors import \
    CommandError, InvalidInputError, UnsupportedError, ConfigurationError, ExternalError, UserError, ProgramError
from .heraldevent import HeraldEvent
from .keyboardkey import KeyboardKey

__all__ = [
    "Command",
    "CommandData",
    "CommandArgs",
    "CommandError",
    "InvalidInputError",
    "UnsupportedError",
    "ConfigurationError",
    "ExternalError",
    "UserError",
    "ProgramError",
    "HeraldEvent",
    "KeyboardKey",
    "ConfigDict",
]



"""The subpackage providing all functions and classes related to databases and tables.

It requires either the ``alchemy_easy`` or the ``alchemy_hard`` extras to be installed.

You can install ``alchemy_easy`` with: ::

    pip install royalnet[alchemy_easy]

To install ``alchemy_hard``, refer to the `psycopg2 <https://pypi.org/project/psycopg2/>}`_ installation instructions,
then run: ::

    pip install royalnet[alchemy_hard]

"""

from .alchemy import Alchemy
from .errors import *
from .table_dfs import table_dfs

__all__ = [
    "Alchemy",
    "table_dfs",
    "AlchemyException",
    "TableNotFoundError",
]



class CommandError(Exception):
    """Something went wrong during the execution of this command.

    Display an error message to the user, explaining what went wrong."""

    def __init__(self, message=""):
        self.message = message

    def __repr__(self):
        return f"{self.__class__.__qualname__}({repr(self.message)})"


class UserError(CommandError):
    """The command failed to execute, and the error is because of something that the user did."""


class InvalidInputError(UserError):
    """The command has received invalid input and cannot complete."""


class UnsupportedError(CommandError):
    """A requested feature is not available on this interface."""


class ConfigurationError(CommandError):
    """The command cannot work because of a wrong configuration by part of the Royalnet admin."""


class ExternalError(CommandError):
    """The command failed to execute, but the problem was because of an external factor (such as an external API going
    down)."""


class ProgramError(CommandError):
    """The command encountered an error in the program."""



import logging
import re
from abc import *
from json import JSONDecodeError
from typing import *

from starlette.requests import Request
from starlette.responses import JSONResponse

import royalnet.utils as ru
from .apidata import ApiData
from .apierrors import *
from .jsonapi import api_error, api_success
from ..pagestar import PageStar

log = logging.getLogger(__name__)


class ApiStar(PageStar, ABC):
    parameters: Dict[str, Dict[str, str]] = {}
    auth: Dict[str, bool] = {}
    deprecated: Dict[str, bool] = {}

    tags: List[str] = []
    __override__: List[str] = []

    async def page(self, request: Request) -> JSONResponse:
        if request.query_params:
            data = request.query_params
        else:
            try:
                data = await request.json()
            except JSONDecodeError:
                data = {}
        apidata = ApiData(data=data, star=self)

        method = request.method.lower()

        try:
            if method == "get":
                response = await self.get(apidata)
            elif method == "post":
                response = await self.post(apidata)
            elif method == "put":
                response = await self.put(apidata)
            elif method == "delete":
                response = await self.delete(apidata)
            elif method == "options":
                return api_success("Preflight allowed.", methods=self.methods())
            else:
                raise MethodNotImplementedError("Unknown method")
        except UnauthorizedError as e:
            return api_error(e, code=401, methods=self.methods())
        except NotFoundError as e:
            return api_error(e, code=404, methods=self.methods())
        except ForbiddenError as e:
            return api_error(e, code=403, methods=self.methods())
        except MethodNotImplementedError as e:
            return api_error(e, code=405, methods=self.methods())
        except BadRequestError as e:
            return api_error(e, code=400, methods=self.methods())
        except Exception as e:
            ru.sentry_exc(e)
            return api_error(e, code=500, methods=self.methods())
        else:
            return api_success(response, methods=self.methods())
        finally:
            await apidata.session_close()

    async def get(self, data: ApiData) -> ru.JSON:
        raise MethodNotImplementedError("GET is not implemented on this ApiStar")

    async def post(self, data: ApiData) -> ru.JSON:
        raise MethodNotImplementedError("POST is not implemented on this ApiStar")

    async def put(self, data: ApiData) -> ru.JSON:
        raise MethodNotImplementedError("PUT is not implemented on this ApiStar")

    async def delete(self, data: ApiData) -> ru.JSON:
        raise MethodNotImplementedError("DELETE is not implemented on this ApiStar")

    def __swagger_for_a_method(self, method: Callable) -> ru.JSON:
        docstring = method.__doc__ or ""
        if docstring is None:
            log.error("Python was started with -OO, so docstrings are disabled and a summary can't be generated.")
            summary = ""
            description = ""
        else:
            summary, description = re.match(r"^(.*)(?:\n{2,}((?:.|\n)*))?", docstring).groups()

        return {
            "operationId": f"{self.__class__.__name__}_{method.__name__}",
            "summary": ru.strip_tabs(summary) if summary is not None else "",
            "description": ru.strip_tabs(description) if description is not None else "",
            "tags": self.tags,
            "deprecated": self.deprecated.get(method.__name__, False),
            "security": [{"RoyalnetLoginToken": ["logged_in"]}] if self.auth.get(method.__name__) else [],
            "parameters": [{
                "name": parameter_name,
                "in": "query",
                "description": ru.strip_tabs(self.parameters[method.__name__][parameter_name]),
                "schema": {},
            } for parameter_name in self.parameters.get(method.__name__, [])],
            "responses": {
                "200": {"description": "✅ OK!"},
                "400": {"description": "⚠️ Missing or invalid parameter."},
                "401": {"description": "⚠️ Invalid password."},
                "403": {"description": "⚠️ Missing or invalid token."},
                "404": {"description": "⚠️ Not found."},
                "405": {"description": "⚠️ Unsupported method."},
                "500": {"description": "⛔️ Serverside unhandled exception!"},
            }
        }

    @classmethod
    def methods(cls):
        magics = []
        for key, value in cls.__dict__.items():
            attr = value
            try:
                magic = attr.__magic__
            except AttributeError:
                continue
            if magic:
                magics.append(key)
        return [*magics, "options"]

    def swagger(self) -> ru.JSON:
        """Generate one or more swagger paths for this ApiStar."""
        result = {}
        for method in self.methods():
            if method not in ["get", "post", "put", "delete"]:
                continue
            result[method.lower()] = self.__swagger_for_a_method(self.__getattribute__(method.lower()))
        return result

        # result = {}
        # for method in cls.methods:
        #     result[method.lower()] = {
        #         "operationId": cls.__name__,
        #         "summary": cls.summary,
        #         "description": cls.description,
        #         "responses": {
        #             "200": {"description": "Success"},
        #             "400": {"description": "Bad request"},
        #             "403": {"description": "Forbidden"},
        #             "404": {"description": "Not found"},
        #             "500": {"description": "Serverside unhandled exception"},
        #             "501": {"description": "Not yet implemented"}
        #         },
        #         "tags": cls.tags,
        #         "parameters": [{
        #             "name": parameter,
        #             "in": "query",
        #             "description": cls.parameters[parameter],
        #             "type": "string"
        #         } for parameter in cls.parameters]
        #     }
        #     if cls.requires_auth:
        #         result[method.lower()]["security"] = [{"RoyalnetLoginToken": ["logged_in"]}]
        # return result



import datetime

import royalnet.constellation.api as rca
import royalnet.constellation.api.apierrors as rcae
import royalnet.utils as ru
from ..tables.tokens import Token
from ..tables.users import User


class ApiAuthLoginRoyalnetStar(rca.ApiStar):
    path = "/api/auth/login/royalnet/v1"

    parameters = {
        "post": {
            "username": "The name of the user you are logging in as.",
            "password": "The password of the user you are logging in as.",
        }
    }

    tags = ["auth"]

    @rca.magic
    async def post(self, data: rca.ApiData) -> ru.JSON:
        """Login with a Royalnet account.

        The method returns a API token valid for 7 days that identifies and authenticates the user to the API.

        Keep it secret, as it is basically a password!
        """
        TokenT = self.alchemy.get(Token)
        UserT = self.alchemy.get(User)

        username = data["username"]
        password = data["password"]

        async with self.session_acm() as session:
            user: User = await ru.asyncify(session.query(UserT).filter_by(username=username).one_or_none)
            if user is None:
                raise rcae.NotFoundError("User not found")
            pswd_check = user.test_password(password)
            if not pswd_check:
                raise rcae.UnauthorizedError("Invalid password")
            token: Token = TokenT.generate(alchemy=self.alchemy, user=user, expiration_delta=datetime.timedelta(days=7))
            session.add(token)
            await ru.asyncify(session.commit)
            response = token.json()

        return response



import abc
import asyncio as aio
import importlib
import logging
import sys
import traceback
from typing import *

from sqlalchemy.schema import Table

import royalnet.alchemy as ra
import royalnet.backpack.tables as rbt
import royalnet.commands as rc
import royalnet.utils as ru

try:
    import royalnet.herald as rh
except ImportError:
    rh = None


log = logging.getLogger(__name__)


class Serf(abc.ABC):
    """An abstract class, to be used as base to implement Royalnet bots on multiple interfaces (such as Telegram or
    Discord)."""
    interface_name = NotImplemented
    prefix = NotImplemented

    _master_table: type = rbt.User
    _identity_table: type = NotImplemented
    _identity_column: str = NotImplemented

    def __init__(self,
                 loop: aio.AbstractEventLoop,
                 alchemy_cfg: rc.ConfigDict,
                 herald_cfg: rc.ConfigDict,
                 packs_cfg: rc.ConfigDict,
                 **_):
        self.loop: Optional[aio.AbstractEventLoop] = loop
        """The event loop this Serf is running on."""

        self.tasks: Optional[ru.TaskList] = ru.TaskList(self.loop)
        """A list of all running tasks of the serf. Initialized at the serf start."""

        # Import packs
        pack_names = packs_cfg["active"]
        packs = {}
        for pack_name in pack_names:
            log.debug(f"Importing pack: {pack_name}")
            try:
                packs[pack_name] = {
                    "commands": importlib.import_module(f".commands", pack_name),
                    "events": importlib.import_module(f".events", pack_name),
                    "stars": importlib.import_module(f".stars", pack_name),
                    "tables": importlib.import_module(f".tables", pack_name),
                }
            except ImportError as e:
                log.error(f"{e.__class__.__name__} during the import of {pack_name}:\n"
                          f"{''.join(traceback.format_exception(*sys.exc_info()))}")
        log.info(f"Packs: {len(packs)} imported")

        self.alchemy: Optional[ra.Alchemy] = None
        """The :class:`Alchemy` object connecting this :class:`Serf` to a database."""

        self.master_table: Optional[Table] = None
        """The central table listing all users. It usually is :class:`User`."""

        self.identity_table: Optional[Table] = None
        """The identity table containing the interface data (such as the Telegram user data) and that is in a 
        many-to-one relationship with the master table."""

        self.identity_column: Optional[str] = None
        """The name of the column in the identity table that contains a unique user identifier. (???)"""

        # Alchemy
        if ra.Alchemy is None:
            log.info("Alchemy: not installed")
        elif not alchemy_cfg["enabled"]:
            log.info("Alchemy: disabled")
        else:
            # Find all tables
            tables = set()
            for pack in packs.values():
                try:
                    # noinspection PyUnresolvedReferences
                    tables = tables.union(pack["tables"].available_tables)
                except AttributeError:
                    log.warning(f"Pack `{pack}` does not have the `available_tables` attribute.")
                    continue
            # Create the Alchemy
            self.init_alchemy(alchemy_cfg, tables)
            log.info(f"Alchemy: {self.alchemy}")

        self.herald: Optional["rh.Link"] = None
        """The :class:`Link` object connecting the :class:`Serf` to the rest of the Herald network."""

        self.herald_task: Optional[aio.Task] = None
        """A reference to the :class:`asyncio.Task` that runs the :class:`Link`."""

        self.events: Dict[str, rc.HeraldEvent] = {}
        """A dictionary containing all :class:`Event` that can be handled by this :class:`Serf`."""

        self.commands: Dict[str, rc.Command] = {}
        """The :class:`dict` connecting each command name to its :class:`Command` object."""

        for pack_name in packs:
            pack = packs[pack_name]
            pack_cfg = packs_cfg.get(pack_name, {})
            try:
                # noinspection PyUnresolvedReferences
                events = pack["events"].available_events
            except AttributeError:
                log.warning(f"Pack `{pack}` does not have the `available_events` attribute.")
            else:
                self.register_events(events, pack_cfg)
            try:
                # noinspection PyUnresolvedReferences
                commands = pack["commands"].available_commands
            except AttributeError:
                log.warning(f"Pack `{pack}` does not have the `available_commands` attribute.")
            else:
                self.register_commands(commands, pack_cfg)
        log.info(f"Events: {len(self.events)} events")
        log.info(f"Commands: {len(self.commands)} commands")

        if rh is None:
            log.info("Herald: not installed")
        elif not herald_cfg["enabled"]:
            log.info("Herald: disabled")
        else:
            self.init_herald(herald_cfg)
            log.info(f"Herald: enabled")

    def init_alchemy(self, alchemy_cfg: Dict[str, Any], tables: Set[type]) -> None:
        """Create and initialize the :class:`Alchemy` with the required tables, and find the link between the master
        table and the identity table."""
        self.alchemy = ra.Alchemy(alchemy_cfg["database_url"], tables)

    @property
    def identity_chain(self) -> tuple:
        """Find a relationship path starting from the master table and ending at the identity table, and return it."""
        return ra.table_dfs(self.master_table, self.identity_table)

    async def call_herald_event(self, destination: str, event_name: str, **kwargs) -> Dict:
        """Send a :class:`royalherald.Request` to a specific destination, and wait for a
        :class:`royalherald.Response`."""
        if self.herald is None:
            raise rc.UnsupportedError("`royalherald` is not enabled on this serf.")
        request: "rh.Request" = rh.Request(handler=event_name, data=kwargs)
        response: "rh.Response" = await self.herald.request(destination=destination, request=request)
        if isinstance(response, rh.ResponseFailure):
            if response.name == "no_event":
                raise rc.ProgramError(f"There is no event named {event_name} in {destination}.")
            elif response.name == "error_in_event":
                if response.extra_info["type"] == "CommandError":
                    raise rc.CommandError(response.extra_info["message"])
                elif response.extra_info["type"] == "UserError":
                    raise rc.UserError(response.extra_info["message"])
                elif response.extra_info["type"] == "InvalidInputError":
                    raise rc.InvalidInputError(response.extra_info["message"])
                elif response.extra_info["type"] == "UnsupportedError":
                    raise rc.UnsupportedError(response.extra_info["message"])
                elif response.extra_info["type"] == "ConfigurationError":
                    raise rc.ConfigurationError(response.extra_info["message"])
                elif response.extra_info["type"] == "ExternalError":
                    raise rc.ExternalError(response.extra_info["message"])
                else:
                    raise rc.ProgramError(f"Invalid error in Herald event '{event_name}':\n"
                                          f"[b]{response.extra_info['type']}[/b]\n"
                                          f"{response.extra_info['message']}")
            elif response.name == "unhandled_exception_in_event":
                raise rc.ProgramError(f"Unhandled exception in Herald event '{event_name}':\n"
                                      f"[b]{response.extra_info['type']}[/b]\n"
                                      f"{response.extra_info['message']}")
            else:
                raise rc.ProgramError(f"Unknown response in Herald event '{event_name}':\n"
                                      f"[b]{response.name}[/b]"
                                      f"[p]{response}[/p]")
        elif isinstance(response, rh.ResponseSuccess):
            return response.data
        else:
            raise rc.ProgramError(f"Other Herald Link returned unknown response:\n"
                                  f"[p]{response}[/p]")

    def register_commands(self, commands: List[Type[rc.Command]], pack_cfg: rc.ConfigDict) -> None:
        """Initialize and register all commands passed as argument."""
        # Instantiate the Commands
        for SelectedCommand in commands:
            # Try to instantiate the command
            try:
                command = SelectedCommand(serf=self, config=pack_cfg)
            except Exception as e:
                log.error(f"Skipping: "
                          f"{SelectedCommand.__qualname__} - {e.__class__.__qualname__} in the initialization.")
                ru.sentry_exc(e)
                continue
            # Warn if the command would be overriding something
            if SelectedCommand.name in self.commands:
                log.info(f"Overriding (already defined): "
                         f"{SelectedCommand.__qualname__} -> {SelectedCommand.name}")
            else:
                log.debug(f"Registering: "
                          f"{SelectedCommand.__qualname__} -> {SelectedCommand.name}")
            # Register the command in the commands dict
            self.commands[SelectedCommand.name] = command
            # Register aliases, but don't override anything
            for alias in SelectedCommand.aliases:
                if alias not in self.commands:
                    log.debug(f"Aliasing: {SelectedCommand.__qualname__} -> {alias}")
                    self.commands[alias] = self.commands[SelectedCommand.name]
                else:
                    log.warning(f"Ignoring (already defined): {SelectedCommand.__qualname__} -> {alias}")

    def init_herald(self, herald_cfg: rc.ConfigDict):
        """Create a :class:`Link` and bind :class:`Event`."""
        herald_cfg["name"] = self.interface_name
        self.herald: "rh.Link" = rh.Link(rh.Config.from_config(**herald_cfg), self.network_handler)

    def register_events(self, events: List[Type[rc.HeraldEvent]], pack_cfg: rc.ConfigDict):
        for SelectedEvent in events:
            # Initialize the event
            try:
                event = SelectedEvent(parent=self, config=pack_cfg)
            except Exception as e:
                log.error(f"Skipping: "
                          f"{SelectedEvent.__qualname__} - {e.__class__.__qualname__} in the initialization.")
                ru.sentry_exc(e)
                continue
            # Register the event
            if SelectedEvent.name in self.events:
                log.warning(f"Overriding (already defined): {SelectedEvent.__qualname__} -> {SelectedEvent.name}")
            else:
                log.debug(f"Registering: {SelectedEvent.__qualname__} -> {SelectedEvent.name}")
            self.events[SelectedEvent.name] = event

    async def network_handler(self, message: Union["rh.Request", "rh.Broadcast"]) -> "rh.Response":
        try:
            event: rc.HeraldEvent = self.events[message.handler]
        except KeyError:
            log.warning(f"No event for '{message.handler}'")
            return rh.ResponseFailure("no_event", f"This serf does not have any event for {message.handler}.")
        log.debug(f"Event called: {event.name}")
        if isinstance(message, rh.Request):
            try:
                response_data = await event.run(**message.data)
                return rh.ResponseSuccess(data=response_data)
            except rc.CommandError as e:
                return rh.ResponseFailure("error_in_event",
                                          f"The event '{message.handler}' raised a {e.__class__.__qualname__}.",
                                          extra_info={
                                              "type": e.__class__.__qualname__,
                                              "message": str(e)
                                          })
            except Exception as e:
                ru.sentry_exc(e)
                return rh.ResponseFailure("unhandled_exception_in_event",
                                          f"The event '{message.handler}' raised an unhandled"
                                          f" {e.__class__.__qualname__}.",
                                          extra_info={
                                              "type": e.__class__.__qualname__,
                                              "message": str(e)
                                          })
        elif isinstance(message, rh.Broadcast):
            await event.run(**message.data)

    async def call(self, command: rc.Command, data: rc.CommandData, parameters: List[str]):
        log.info(f"Calling command: {command.name}")
        try:
            # Run the command
            await command.run(rc.CommandArgs(parameters), data)
        except rc.InvalidInputError as e:
            await data.reply(f"⚠️ {e.message}\n"
                             f"Syntax: [c]{self.prefix}{command.name} {command.syntax}[/c]")
        except rc.UserError as e:
            await data.reply(f"⚠️ {e.message}")
        except rc.UnsupportedError as e:
            await data.reply(f"⚠️ {e.message}")
        except rc.ExternalError as e:
            await data.reply(f"⚠️ {e.message}")
        except rc.ConfigurationError as e:
            await data.reply(f"⚠️ {e.message}")
        except rc.ProgramError as e:
            await data.reply(f"⛔️ {e.message}")
        except rc.CommandError as e:
            await data.reply(f"⚠️ {e.message}")
        except Exception as e:
            ru.sentry_exc(e)
            await data.reply(f"⛔️ [b]{e.__class__.__name__}[/b]\n" + '\n'.join(map(lambda a: repr(a), e.args)))

    @staticmethod
    async def press(key: rc.KeyboardKey, data: rc.CommandData):
        log.info(f"Calling key_callback: {repr(key)}")
        try:
            await key.press(data)
        except rc.InvalidInputError as e:
            await data.reply(f"⚠️ {e.message}")
        except rc.UserError as e:
            await data.reply(f"⚠️ {e.message}")
        except rc.UnsupportedError as e:
            await data.reply(f"⚠️ {e.message}")
        except rc.ExternalError as e:
            await data.reply(f"⚠️ {e.message}")
        except rc.ConfigurationError as e:
            await data.reply(f"⚠️ {e.message}")
        except rc.ProgramError as e:
            await data.reply(f"⛔️ {e.message}")
        except rc.CommandError as e:
            await data.reply(f"⚠️ {e.message}")
        except Exception as e:
            ru.sentry_exc(e)
            await data.reply(f"⛔️ [b]{e.__class__.__name__}[/b]\n" + '\n'.join(map(lambda a: repr(a), e.args)))

    async def run(self):
        """A coroutine that starts the event loop and handles command calls."""
        if self.herald is not None:
            self.herald_task = self.tasks.add(self.herald.run())
        # OVERRIDE THIS METHOD!

    @classmethod
    def run_process(cls, **kwargs):
        """Blockingly create and run the Serf.

        This should be used as the target of a :class:`multiprocessing.Process`."""
        ru.init_logging(kwargs["logging_cfg"])

        if kwargs["sentry_cfg"] is None or not kwargs["sentry_cfg"]["enabled"]:
            log.info("Sentry: disabled")
        else:
            try:
                ru.init_sentry(kwargs["sentry_cfg"])
            except ImportError:
                log.info("Sentry: not installed")

        loop = aio.get_event_loop()

        serf = cls(loop=loop, **kwargs)

        try:
            serf.loop.run_until_complete(serf.run())
        except Exception as e:
            ru.sentry_exc(e, level="fatal")



import asyncio as aio
import io
import logging
import sys
from typing import *

import discord

import royalnet.backpack.tables as rbt
import royalnet.commands as rc
from royalnet.serf import Serf
from royalnet.utils import asyncify, sentry_exc
from .escape import escape

log = logging.getLogger(__name__)


class DiscordSerf(Serf):
    """A :class:`Serf` that connects to `Discord <https://discordapp.com/>`_ as a bot."""
    interface_name = "discord"
    prefix = "!"

    def __init__(self,
                 loop: aio.AbstractEventLoop,
                 alchemy_cfg: rc.ConfigDict,
                 herald_cfg: rc.ConfigDict,
                 sentry_cfg: rc.ConfigDict,
                 packs_cfg: rc.ConfigDict,
                 serf_cfg: rc.ConfigDict,
                 **_):
        if discord is None:
            raise ImportError("'discord' extra is not installed")

        super().__init__(loop=loop,
                         alchemy_cfg=alchemy_cfg,
                         herald_cfg=herald_cfg,
                         sentry_cfg=sentry_cfg,
                         packs_cfg=packs_cfg,
                         serf_cfg=serf_cfg)

        self.token = serf_cfg["token"]
        """The Discord bot token."""

        self.Client = self.client_factory()
        """The custom :class:`discord.Client` class that will be instantiated later."""

        self.client = self.Client(status=discord.Status.do_not_disturb)
        """The custom :class:`discord.Client` instance."""

        self.Data: Type[rc.CommandData] = self.data_factory()

    def data_factory(self) -> Type[rc.CommandData]:
        # noinspection PyMethodParameters,PyAbstractClass
        class DiscordData(rc.CommandData):
            def __init__(data,
                         command: rc.Command,
                         message: "discord.Message"):
                super().__init__(command=command)
                data.message: "discord.Message" = message

            async def reply(data, text: str):
                await data.message.channel.send(escape(text))

            async def reply_image(data, image: io.IOBase, caption: Optional[str] = None) -> None:
                await data.message.channel.send(caption, file=discord.File(image, 'image'))

            async def find_author(data,
                                  *,
                                  session,
                                  required: bool = False) -> Optional[rbt.User]:
                user: Union["discord.User", "discord.Member"] = data.message.author
                DiscordT = data.alchemy.get(rbt.Discord)
                result = await asyncify(
                    session.query(DiscordT).filter(DiscordT.discord_id == user.id).one_or_none
                )
                if result is None and required:
                    raise rc.CommandError("You must be registered to use this command.")
                return result

            async def delete_invoking(data, error_if_unavailable=False):
                await data.message.delete()

        return DiscordData

    async def handle_message(self, message: "discord.Message"):
        """Handle a Discord message by calling a command if appropriate."""
        text = message.content
        # Skip non-text messages
        if not text:
            log.debug("Skipping message as it had no text")
            return
        # Skip non-command updates
        if not text.startswith(self.prefix):
            log.debug(f"Skipping message as it didn't start with {self.prefix}")
            return
        # Skip bot messages
        author: Union["discord.User", "discord.Member"] = message.author
        if author.bot:
            log.debug(f"Skipping message as it was from another bot")
            return
        # Find and clean parameters
        command_text, *parameters = text.split(" ")
        # Don't use a case-sensitive command name
        command_name = command_text.lstrip(self.prefix).lower()
        log.debug(f"Parsed '{command_name}' as command name")
        # Find the command
        try:
            command = self.commands[command_name]
        except KeyError:
            # Skip the message
            log.debug(f"Skipping message as I could not find the command {command_name}")
            return
        # Call the command
        log.debug(f"Sending typing notification")
        with message.channel.typing():
            # Prepare data
            # noinspection PyArgumentList
            data = self.Data(command=command, message=message)
            # Call the command
            log.debug(f"Calling {command}")
            await self.call(command, data, parameters)

    def client_factory(self) -> Type["discord.Client"]:
        """Create a custom class inheriting from :py:class:`discord.Client`."""

        # noinspection PyMethodParameters
        class DiscordClient(discord.Client):
            # noinspection PyMethodMayBeStatic
            async def on_message(cli, message: "discord.Message") -> None:
                """Handle messages received by passing them to the handle_message method of the bot."""
                self.tasks.add(self.handle_message(message))

            async def on_ready(cli) -> None:
                """Change the bot presence to ``online`` when the bot is ready."""
                log.debug("Discord client is ready!")
                await cli.change_presence(status=discord.Status.online, activity=None)

            # noinspection PyMethodMayBeStatic
            async def on_resume(cli) -> None:
                log.debug("Discord client resumed connection.")

            async def on_error(self, event_method, *args, **kwargs):
                exc_type, exc_obj, exc_tb = sys.exc_info()
                sentry_exc(exc_obj)

        return DiscordClient

    async def run(self):
        await super().run()
        await self.client.login(self.token)
        while True:
            try:
                await self.client.connect(reconnect=False)
            except discord.GatewayNotFound:
                log.error("Discord Gateway not found! Retrying in 60 seconds...")
                await aio.sleep(60)
            except discord.ConnectionClosed:
                log.error("Discord connection was closed! Retrying in 15 seconds...")
                await aio.sleep(60)



import functools


def magic(func):
    """Mark a function as "magic", so its override can be detected externally."""
    func.__magic__ = True

    @functools.wraps(func)
    async def f(*args, **kwargs):
        return await func(*args, **kwargs)

    return f



import royalnet.backpack.tables as rbt
import royalnet.constellation.api as rca
import royalnet.utils as ru


class ApiUserGetStar(rca.ApiStar):
    path = "/api/user/get/v1"

    parameters = {
        "get": {
            "id": "The id of the user to get."
        }
    }

    tags = ["user"]

    @rca.magic
    async def get(self, data: rca.ApiData) -> dict:
        """Get details about the Royalnet user with a certain id."""
        user_id_str = data["id"]
        try:
            user_id = int(user_id_str)
        except (ValueError, TypeError):
            raise rca.InvalidParameterError("'id' is not a valid int.")
        user: rbt.User = await ru.asyncify(data.session.query(self.alchemy.get(rbt.User)).get, user_id)
        if user is None:
            raise rca.NotFoundError("No such user.")
        return user.json()



import re
from typing import *


def escape(string: Optional[str]) -> Optional[str]:
    """Escape a string to be sent through Telegram (as HTML), and format it using RoyalCode.

    Warning:
        Currently escapes everything, even items in code blocks."""

    url_pattern = re.compile(r"\[url=(.*?)](.*?)\[/url]")
    url_replacement = r'<a href="\1">\2</a>'

    escaped_string = string.replace("<", "&lt;").replace(">", "&gt;")

    simple_parse = escaped_string \
        .replace("[b]", "<b>") \
        .replace("[/b]", "</b>") \
        .replace("[i]", "<i>") \
        .replace("[/i]", "</i>") \
        .replace("[u]", "<b>") \
        .replace("[/u]", "</b>") \
        .replace("[c]", "<code>") \
        .replace("[/c]", "</code>") \
        .replace("[p]", "<pre>") \
        .replace("[/p]", "</pre>")

    advanced_parse = re.sub(url_pattern, url_replacement, simple_parse)

    return advanced_parse



import logging
from asyncio import Event
from contextlib import asynccontextmanager

log = logging.getLogger(__name__)


class MultiLock:
    """A lock that allows either simultaneous read access or exclusive write access.

    Basically, a reimplementation of Rust's `RwLock <https://doc.rust-lang.org/beta/std/sync/struct.RwLock.html>`_ ."""

    def __init__(self):
        self._counter: int = 0
        self._normal_event: Event = Event()
        self._exclusive_event: Event = Event()
        self._normal_event.set()
        self._exclusive_event.set()

    def _check_event(self):
        if self._counter > 0:
            self._normal_event.clear()
        else:
            self._normal_event.set()

    @asynccontextmanager
    async def normal(self):
        """Acquire the lock for simultaneous access."""
        log.debug(f"Waiting for exclusive lock end: {self}")
        await self._exclusive_event.wait()
        log.debug(f"Acquiring normal lock: {self}")
        self._counter += 1
        self._check_event()
        try:
            yield
        finally:
            log.debug(f"Releasing normal lock: {self}")
            self._counter -= 1
            self._check_event()

    @asynccontextmanager
    async def exclusive(self):
        """Acquire the lock for exclusive access."""
        log.debug(f"Waiting for exclusive lock end: {self}")
        await self._exclusive_event.wait()
        self._exclusive_event.clear()
        log.debug(f"Waiting for normal lock end: {self}")
        await self._normal_event.wait()
        try:
            log.debug(f"Acquiring exclusive lock: {self}")
            self._exclusive_event.clear()
            yield
        finally:
            log.debug(f"Releasing exclusive lock: {self}")
            self._exclusive_event.set()

    def __repr__(self):
        return f"<MultiLock {self._counter} " \
               f"{'_' if self._normal_event.is_set() else 'N'}{'_' if self._exclusive_event.is_set() else 'E'}>"



import base64
import uuid as _uuid


def to_urluuid(uuid: _uuid.UUID) -> str:
    """Return a base64 url-friendly short UUID."""
    return str(base64.urlsafe_b64encode(uuid.bytes), encoding="ascii").rstrip("=")


def from_urluuid(b: str) -> _uuid.UUID:
    return _uuid.UUID(bytes=base64.urlsafe_b64decode(bytes(b + "==", encoding="ascii")))



import asyncio as aio
import logging
from typing import *

from .sentry import sentry_exc

log = logging.getLogger(__name__)


class TaskList:
    def __init__(self, loop: aio.AbstractEventLoop):
        self.loop: aio.AbstractEventLoop = loop
        self.tasks: List[aio.Task] = []

    def collect(self):
        """Remove finished tasks from the list."""
        log.debug(f"Collecting done tasks")
        new_list = []
        for task in self.tasks:
            try:
                task.result()
            except aio.CancelledError:
                log.warning(f"Task {task} was unexpectedly cancelled.")
            except aio.InvalidStateError:
                log.debug(f"Task {task} hasn't finished running yet, readding it to the list.")
                new_list.append(task)
            except Exception as err:
                sentry_exc(err)
        self.tasks = new_list

    def add(self, coroutine: Awaitable[Any], timeout: float = None) -> aio.Task:
        """Add a new task to the list; the task will be cancelled if ``timeout`` seconds pass."""
        log.debug(f"Creating new task {coroutine}")
        if timeout:
            task = self.loop.create_task(aio.wait_for(coroutine, timeout=timeout))
        else:
            task = self.loop.create_task(coroutine)
        self.tasks.append(task)
        return task



# Imports go here!
from .aliases import Alias
from .discord import Discord
from .roles import Role
from .telegram import Telegram
from .tokens import Token
from .users import User

# Enter the tables of your Pack here!
available_tables = {
    User,
    Telegram,
    Discord,
    Alias,
    Token,
    Role,
}

# Don't change this, it should automatically generate __all__
__all__ = [table.__name__ for table in available_tables]



import logging
from typing import *

try:
    import coloredlogs
except ImportError:
    coloredlogs = None

l: logging.Logger = logging.getLogger(__name__)


# From https://stackoverflow.com/a/56810619/4334568
def reset_logging():
    manager = logging.root.manager
    manager.disabled = logging.NOTSET
    for logger in manager.loggerDict.values():
        if isinstance(logger, logging.Logger):
            logger.setLevel(logging.NOTSET)
            logger.propagate = True
            logger.disabled = False
            logger.filters.clear()
            handlers = logger.handlers.copy()
            for handler in handlers:
                # Copied from `logging.shutdown`.
                try:
                    handler.acquire()
                    handler.flush()
                    handler.close()
                except (OSError, ValueError):
                    pass
                finally:
                    handler.release()
                logger.removeHandler(handler)


def init_logging(logging_cfg: Dict[str, Any]):
    reset_logging()
    loggers_cfg = logging_cfg["Loggers"]
    for logger_name in loggers_cfg:
        if logger_name == "root":
            log: logging.Logger = logging.root
        else:
            log: logging.Logger = logging.getLogger(logger_name)
        log.setLevel(loggers_cfg[logger_name])

    stream_handler = logging.StreamHandler()
    if coloredlogs is not None:
        stream_handler.formatter = coloredlogs.ColoredFormatter(logging_cfg["log_format"], style="{")
    else:
        stream_handler.formatter = logging.Formatter(logging_cfg["log_format"], style="{")
    logging.root.handlers.clear()
    logging.root.addHandler(stream_handler)

    l.debug("Logging: ready")
